import {
  ConflictException,
  ForbiddenException,
  Injectable,
  NotFoundException,
} from '@nestjs/common';
import { InjectRepository } from '@nestjs/typeorm';
import { Repository } from 'typeorm';
import { InstitucionTipoCarrito } from './entity/institucion-tipo-carrito.entity';
import { Operador } from '../operador/entity/operador.entity';
import { TipoCarrito } from './entity/tipo-carrito.entity';
import { Usuario } from '../usuario/entity/usuario.entity';
import { InstitucionService } from '../institucion/institucion.service';
import { InstitucionUsuarioService } from '../institucion-usuario/institucion-usuario.service';

@Injectable()
export class InstitucionTipoCarritoService {
  constructor(
    @InjectRepository(TipoCarrito)
    private tipoCarritoRepository: Repository<TipoCarrito>,
    @InjectRepository(InstitucionTipoCarrito)
    private institucionTipoCarritoRepository: Repository<InstitucionTipoCarrito>,
    private institucionService: InstitucionService,
    private institucionUsuarioService: InstitucionUsuarioService,
  ) {}

  create(letra: string, tipo_carrito: string): Promise<{ message: string }> {
    // Buscamos un registro con ese nombre
    return this.tipoCarritoRepository
      .findOne({ where: [{ letra }, { tipo_carrito }] })
      .then((existeTipoCarrito) => {
        // Sacamos error si existe
        if (existeTipoCarrito) {
          if (letra === existeTipoCarrito.letra)
            throw new ConflictException(
              'Ya existe un tipo de carrito con esta letra, intente con otra.',
            );
          else
            throw new ConflictException(
              'Ya existe un tipo carrito con este nombre, intente con otro.',
            );
        }
        // Creamos y guardamos un registro
        return this.tipoCarritoRepository.save(
          this.tipoCarritoRepository.create({ letra, tipo_carrito }),
        );
      })
      .then(async (tipoCarrito) => {
        const instituciones = await this.institucionService.findAll();

        // Asociamos el nuevo tipo de carrito con las instituciones existentes
        for (let i = 0; i < instituciones.length; i++)
          await this.institucionTipoCarritoRepository.save(
            this.institucionTipoCarritoRepository.create({
              institucion: instituciones[i],
              tipoCarrito,
            }),
          );
        return { message: 'Se creó correctamente un nuevo tipo de carrito.' };
      });
  }

  findAll(): Promise<TipoCarrito[]> {
    return this.tipoCarritoRepository.find({ order: { tipo_carrito: 'ASC' } });
  }

  async findAllMostrar(
    usuario: Usuario,
    id_institucion: number,
  ): Promise<InstitucionTipoCarrito[]> {
    const institucion = await this.institucionService.findById(id_institucion);
    const institucionUsuario =
      await this.institucionUsuarioService.findByIdUsuarioIdInstitucion(
        usuario,
        institucion,
      );

    if (!institucionUsuario)
      throw new ForbiddenException(
        'No puedes acceder a esta información porque no peteneces a esta institución.',
      );
    return this.institucionTipoCarritoRepository.find({
      where: { mostrar: true, institucion },
      order: { tipoCarrito: { tipo_carrito: 'ASC' } },
    });
  }

  findAllOperador(operador: Operador): Promise<InstitucionTipoCarrito[]> {
    return this.institucionTipoCarritoRepository.find({
      where: { institucion: operador.institucion },
      order: { tipoCarrito: { tipo_carrito: 'ASC' } },
    });
  }

  findById(id_institucion_tipo_carrito): Promise<InstitucionTipoCarrito> {
    return this.institucionTipoCarritoRepository
      .findOne({ where: { id_institucion_tipo_carrito } })
      .then((institucionTipoCarrito) => {
        if (!institucionTipoCarrito)
          throw new NotFoundException(
            'No existe este id institución tipo carrito.',
          );
        return institucionTipoCarrito;
      });
  }

  findTipoCarritoById(id_tipo_carrito: number): Promise<TipoCarrito> {
    return this.tipoCarritoRepository
      .findOne({ where: { id_tipo_carrito } })
      .then((tipoCarrito) => {
        if (!tipoCarrito)
          throw new NotFoundException('No existe este id tipo carrito.');
        return tipoCarrito;
      });
  }

  findTipoCarritoByTipoCarrito(tipo_carrito: string): Promise<TipoCarrito> {
    return this.tipoCarritoRepository.findOne({ where: { tipo_carrito } });
  }

  update(
    admin: Operador,
    attrs: Partial<InstitucionTipoCarrito>,
  ): Promise<{ message: string }> {
    return this.findById(attrs.id_institucion_tipo_carrito)
      .then((institucionTipoCarrito) => {
        // Validamos que la institucion tipo carrito pertenezca a la institución del admin
        if (
          admin.institucion.id_institucion !=
          institucionTipoCarrito.institucion.id_institucion
        )
          throw new ConflictException(
            'No puedes modificar la información este tipo de carrito porque no pertenece a tu institución.',
          );
        // Asignamos valores enviados al objeto
        Object.assign(institucionTipoCarrito, attrs);
        // Guardamos cambios
        return this.institucionTipoCarritoRepository.save(
          institucionTipoCarrito,
        );
      })
      .then((_) => ({ message: 'Se guardaron los cambios correctamente.' }));
  }
}
